#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#ifdef OPEN_BSD
#include <netinet/in_systm.h>
#endif
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/timeb.h>
#include <signal.h>
#include <time.h>
#include <pcap.h>
#include "lib.h"
#if (SOLARIS_SPARC || OPEN_BSD)
#include "gnuip.h"
#endif
#ifdef HAVE_REGEX
#include <regex.h>
#endif
#include "iphash.h"


static struct timeb starttime;
static struct timeb lasttime;
unsigned long bytes;
unsigned long packets;
unsigned long long totalbytes;
unsigned long long totalpackets;

unsigned int opt_delay = 1;
unsigned int opt_prefix = 0;
unsigned int opt_bits = 0;
int opt_counter = -1;
int opt_human = 0;

struct ipclassinfo * ipci = NULL;
int opt_printtop = 10;
int opt_local = 0;
int opt_zero = 0;
int opt_ips = 0;

#ifdef HAVE_REGEX
char * opt_regex = NULL;
char * opt_extractor = NULL;
char opt_nul_char = '@';
#endif

int opt_caplen = 40;

void host_inetntoa(char * str, int max, u_int32_t ip)
{
	snprintf(str, max, "%d.%d.%d.%d", (ip >> 24) & 0xff, (ip >> 16) & 0xff, (ip >> 8) & 0xff, ip & 0xff);
}

void show(struct top * top, char * header, unsigned long millitm)
{
	int n, i, left = 0, right;
	char unit[5];
	
	printf("%s:\n", header);
	n = top->n / 2;
	right = n;
	
	if(opt_bits)
		snprintf(unit, 5, "bps");
	else
		snprintf(unit, 5, "Bps");
	
	for(i = 0; i<n; i++)
	{
		char leftip[16];
		char rightip[16];
		char leftstatus[18];
		char rightstatus[18];
		double lv;
		double rv;
		
		lv = top->top[left].data;
		rv = top->top[right].data;
		lv /= millitm; rv /= millitm;
		lv *= 1000; rv *= 1000;
		if(opt_bits)
		{
			lv *= 8;
			rv *= 8;
		}
		
		host_inetntoa(leftip, 16, top->top[left].ip);
		host_inetntoa(rightip, 16, top->top[right].ip);
		mkhuman(lv, 18, leftstatus, unit, 1024, opt_human);
		mkhuman(rv, 18, rightstatus, unit, 1024, opt_human);
		
		if(!top->top[left].ip) continue;
		if(!top->top[right].ip)
			printf("%3d.%15s %18s\n", left + 1, leftip, leftstatus);
		else 
			printf("%3d.%15s %18s %3d.%15s %18s\n", left + 1, leftip, leftstatus, right + 1, rightip, rightstatus);
		
		left++; right++;
		
	}
}

void abusers_output(int signal)
{
	time_t time;
	unsigned short millitm;
	unsigned long miliFromLast, miliFromStart;
	struct top * intop;
	struct top * outtop;

	alarm(opt_delay);
	
	time = lasttime.time;
	millitm = lasttime.millitm;
	
	ftime(&lasttime);
	
	miliFromStart = (lasttime.time - starttime.time) * 1000 + lasttime.millitm - starttime.millitm;
	miliFromLast = (lasttime.time - time) * 1000 + lasttime.millitm - millitm;

	intop = findtop(opt_printtop, 1);
	outtop = findtop(opt_printtop, 0);

	if(opt_zero)
	{
		show(intop, "Input bandwidth abusers:", miliFromLast);
		show(outtop, "Output bandwidth abusers:", miliFromLast);
	}
	else
	{
		show(intop, "Input bandwidth abusers:", miliFromStart);
		show(outtop, "Output bandwidth abusers:", miliFromStart);
	}
	
	free(intop->top); free(intop);
	free(outtop->top); free(outtop);

	opt_counter--;
	if(opt_counter == 0) exit(0);
	if(opt_zero) zero_entries();
}


void output(int signal)
{
	time_t time;
	unsigned short millitm;
	unsigned long nowbytes, nowpackets;
	unsigned long miliFromLast, miliFromStart;
	double bpsFromStart, bpsFromLast, ppsFromStart, ppsFromLast;
	char bpsFromLastStr[100];
	char bpsFromStartStr[100];
	char ppsFromLastStr[100];
	char ppsFromStartStr[100];

#ifdef HAVE_REGEX
	if(opt_extractor)
		return;
#endif

	alarm(opt_delay);
	
	time = lasttime.time;
	millitm = lasttime.millitm;
	
	nowbytes = bytes; bytes = 0;
	nowpackets = packets; packets = 0;
	ftime(&lasttime);
	
	totalbytes += nowbytes;
	totalpackets += nowpackets;
	
	miliFromStart = (lasttime.time - starttime.time) * 1000 + lasttime.millitm - starttime.millitm;
	miliFromLast = (lasttime.time - time) * 1000 + lasttime.millitm - millitm;
	
	ppsFromStart = totalpackets; ppsFromStart /= miliFromStart;
	ppsFromLast = nowpackets; ppsFromLast /= miliFromLast;
	bpsFromStart = totalbytes; bpsFromStart /= miliFromStart;
	bpsFromLast = nowbytes; bpsFromLast /= miliFromLast;
	ppsFromStart *= 1000; ppsFromLast *= 1000;
	bpsFromStart *= 1000; bpsFromLast *= 1000;
	
	if(opt_bits)
		mkhuman(bpsFromLast * 8, 100, bpsFromLastStr, "bps", 1024, opt_human);
	else
		mkhuman(bpsFromLast, 100, bpsFromLastStr, "Bps", 1024, opt_human);
		
	mkhuman(ppsFromLast, 100, ppsFromLastStr, "pps", 1000, opt_human);
	
	if(opt_bits)
		mkhuman(bpsFromStart * 8, 100, bpsFromStartStr, "bps", 1024, opt_human);
	else
		mkhuman(bpsFromStart, 100, bpsFromStartStr, "Bps", 1024, opt_human);

	mkhuman(ppsFromStart, 100, ppsFromStartStr, "pps", 1000, opt_human);
	printf("-> Currently %s/%s, Average: %s/%s\n", bpsFromLastStr, ppsFromLastStr, bpsFromStartStr, ppsFromStartStr);
	opt_counter--;
	if(opt_counter == 0) exit(0);
}


void help(unsigned char * name)
{
	printf("bmtools 0.5 - Mateusz 'mteg' Golicz <mtg@elsat.net.pl>, 2003\n", name);
	printf("usage: %s [options]\n", name);
	printf("  -h        Show this help\n");
	printf("  -a <t>    Show bandwidth abusers mode: print top <t>\n");
	printf("  -l        Account local transfers too.\n");
	printf("  -c <class>Consider nodes from this class (turns on abusers mode)\n");
	printf("  -u        More human-readable output\n");
	printf("  -b        Show output in bps (not Bps)\n");
	printf("  -p <pref> Datalink opt_prefix (gets substracted from each packet size, default: 0)\n");
	printf("  -i <int>  Use interface <int> - default eth0\n");
	printf("  -f <bpf>  BPF filter expression to use\n");
	printf("  -r <t>    Print results every <t> seconds (default: 1)\n");
	printf("  -q <r>    Quit after printing r results\n");
	printf("  -s <b>    Capture l bytes (default: 256)\n");
#ifdef HAVE_REGEX
	printf("  -x <rege> Match this regex in packet (increases capture length to at least 1500b)\n");
	printf("  -0 n      Replace nul character with this before doing regex matches (default: '@')\n");
	printf("  -e <regex>Extractor mode.\n");
	printf("  -o <forma>Extractor output format.\n");
	printf("  -I        Show IP addresses before extractor output.\n");
#endif
	printf("  -z        Zero counters after each output (calculate current, not average, turns on abusers)\n");
//	printf("  -S <name> Save operation mode.\n");
//	printf("  -L <name> Recall operation mode.\n");
}

void abusers_callback(unsigned char *userdata,const struct pcap_pkthdr *hd,const unsigned char *data)
{
	struct iphdr *iph;
	u_int32_t sip;
	u_int32_t dip;
	u_int32_t host;
	long long in = 0, out = 0;

	if(hd->caplen < (opt_prefix + sizeof(struct iphdr))) return;


	iph=(struct iphdr*) (char*) (((char*) data)+opt_prefix);
			
	if(iph->version != 4) return;

	sip = iph->saddr;
	dip = iph->daddr;

 	if(is_ours(ipci, sip))
        	{ host = sip; out = hd->len - opt_prefix;}
	    
	if(is_ours(ipci, dip))
		{ host = dip; in = hd->len - opt_prefix;	}

	if(!(in || out)) return;
	if(in && out && (!opt_local)) return;

#ifdef HAVE_REGEX
	if(opt_regex)
		if(!check_regex(data + opt_prefix, hd->caplen - opt_prefix, opt_nul_char)) return;
#endif

	add_entry(ntohl(host), in, out);
}

void counter_callback(unsigned char *userdata,const struct pcap_pkthdr *hd,const unsigned char *data)
{
#ifdef HAVE_REGEX
	if(opt_regex)
		if(!check_regex(data + opt_prefix, hd->caplen - opt_prefix, opt_nul_char)) return;

	if(opt_extractor)
	{
		extractor(data + opt_prefix, hd->caplen - opt_prefix, opt_nul_char, opt_ips);
	}
	else
#endif
	{
		bytes += hd->caplen - opt_prefix;
		packets += 1;
	}
}


int main(int argc, char ** argv)
{
	unsigned char * bpfexpression = NULL;
	unsigned char * interface = "eth0";
	char ebuff[PCAP_ERRBUF_SIZE];
	pcap_t *handle;
	struct bpf_program bpf;
	int c, abusers_mode = 0;
	char *opt_order = NULL;
	
	while((c = getopt(argc, argv, "hi:f:r:up:bq:l0zx:s:c:a:e:o:I")) != EOF)
	{
		switch(c)
		{
#if HAVE_REGEX
			case '0':
				opt_nul_char = *optarg;
				break;
			case 'o':
				opt_order = strdup(optarg);
				break;
			case 'e':
				if(opt_caplen < 1500) opt_caplen = 1500;
				opt_extractor = strdup(optarg);
				break;
			case 'x':
				if(opt_caplen < 1500) opt_caplen = 1500;
				opt_regex = strdup(optarg);
				break;
			case 'I':
				opt_ips = 1;
				break;
#endif
			case 'a':
				opt_printtop = atoi(optarg);
				if(opt_printtop < 1) opt_printtop = 1;
				abusers_mode = 1;
				break;
			case 'c':
				ipci = add_class(optarg, ipci);
				if(!ipci) return(1);
				abusers_mode = 1;
				break;
			case 'z': opt_zero = 1; abusers_mode = 1; break;
			case 'l': opt_local = 1; break;
			case 'u': opt_human = 1; break;
			case 'b': opt_bits = 1; break;
			case 'q': opt_counter = atoi(optarg) ; break;
			case 'p': opt_prefix = atoi(optarg); break;
			case 'h':
				help(argv[0]);
				exit(0);
			case 'i':
				interface = strdup(optarg);
				break;
			case 'f':
				bpfexpression = strdup(optarg);
				break;
			case 'r':
				opt_delay = atoi(optarg);
				if(opt_delay < 1) opt_delay = 1;
				break;
		}
	}
	signal(SIGPIPE, SIG_IGN);

	if(abusers_mode)
		signal(SIGALRM, abusers_output);
	else
		signal(SIGALRM, output);
	
	if(argv[optind] && !bpfexpression)
		bpfexpression = argv[optind++];
	
	if(argv[optind])
	{
		fprintf(stderr, "Unrecognized command line parameters (starting from: %s)...\n", argv[optind]);
		exit(1);
	}
	
	if(opt_regex)
		prepare_regex(opt_regex, opt_caplen);

	if(opt_extractor)
		prepare_extractor(opt_extractor, opt_order, opt_caplen);
	
	if((handle = pcap_open_live(interface, 2048, -1, 50, ebuff))==NULL)
	{
		fprintf(stderr, "pcap_open_live(): %s\n", ebuff);
		return(0);
 	}

	if(bpfexpression)
	{
		if(pcap_compile(handle, &bpf, bpfexpression, 1, 0xffffffff) == -1)
		{
			fprintf(stderr, "pcap_compile(): %s\n", pcap_geterr(handle));
			return(1);
		}
		if(pcap_setfilter(handle, &bpf) == -1)
		{
			fprintf(stderr, "pcap_compile(): %s\n", pcap_geterr(handle));
			return(1);
		}
	}

	
	alarm(opt_delay);
	ftime(&starttime);
	ftime(&lasttime);
	if(abusers_mode)
		pcap_loop(handle, -1, abusers_callback, NULL);
	else
		pcap_loop(handle, -1, counter_callback, NULL);
	pcap_close(handle);
}
